"""Constants."""

from __future__ import annotations

__all__ = ('CONTEXT_SETTINGS',)

CONTEXT_SETTINGS = {'help_option_names': ('-h', '--help')}
"""
Shared context settings for all commands.

:meta hide-value:
"""
